#include <p18cxxx.h>
#include <i2c.h>
#include <usart.h>
#include <stdlib.h>

char button = 0;		//scan input
char input = 16;		//button value
char previnput = 16;	//previous input
char i = 0;				//column iterator
const unsigned char sevSeg[17] = {	0x7E, 0x30, 0x6D, 0x79, 0x33, 0x5B, 0x5F, 0x70, 0x7F, 0x7B,		//7 segment codes
									0x77, 0x1F, 0x4e, 0x3d, 0x4f, 0x47 , 0x00 };
const char keymap[4][4] = {	{1,2,3,12},		//column & row to button value
							{4,5,6,13},
							{7,8,9,16},
							{10,0,11,14}};

const char columnMask [4] = { 0x7F, 0xBF, 0xDF, 0xEF };	//column select values

char str[20] = "Input: ";
char end[3] = "\r\n";

void init(void){
	OSCCONbits.IRCF0= 1;
	OSCCONbits.IRCF1= 1;	//set internal clock to 8MHz
	OSCCONbits.IRCF2= 1;

	ADCON1 = 0xFF;	//disable analog inputs	

    TRISB = 0x00;	//set PORTB as output pins
    PORTB = 0x00;	//make all output pins LOW
	OpenI2C( MASTER, SLEW_OFF);	//configure I2C module as MASTER
	SSPADD = 0x27;	//set I2C clock to 100kHz with 8MHz base clock (PIC18F4550 datasheet page 223)

	OpenUSART(	USART_TX_INT_OFF & USART_RX_INT_OFF & USART_ASYNCH_MODE &	//configure UART module in async mode without interupts
				USART_EIGHT_BIT &   USART_CONT_RX & USART_BRGH_LOW, 12); 	//8bit mode and baudrate of 9600Bd (8MHz/(64*(12+1))
}

/*	The keypad is scanned by selecting a column and then reading the rows to see wich one of the buttons is pressed.
	First all the output pins (rows and colums) of the I2C chip are set high except the column we are going to scan, this one is set low.
	After that we set the I2C chip in read mode and read the pins. If one of the buttons on this column is pressed,
	the row pin it's connected to is pulled low because pressing the button connects the row and column pin.
	We can now determine the button value because we have the row and column value.
	If none of the rows is low, we select the next column and read again.
*/
char scan(void){
	for(i = 0; i < 4; i++){		//scan the columns until a pressed button is found
		RestartI2C();		//restart I2C communication
			IdleI2C();		//wait until I2C is ready
		WriteI2C( 0x40 );	//send address to the device (0100 (hardcoded in chip) + 000 (address pins) + 0 (write mode))
			IdleI2C();		//wait until I2C is ready
		WriteI2C( columnMask[i] );	//select one of the columns
			IdleI2C();		//wait until I2C is ready
		RestartI2C();		//begin I2C communication
			IdleI2C();		//wait until I2C is ready
		WriteI2C( 0x41 );	//send address to the device (0100 (hardcoded in chip) + 000 (address pins) + 1 (read mode))
			IdleI2C();		//wait until I2C is ready
		button = ReadI2C();	//read value of rows and columns
			IdleI2C();		//wait until I2C is ready
		button = button | 0xF0;	//set column values high

		switch(button)
		{
			case 0xF7:	return keymap[0][i];	//first row pressed
			case 0xFB:	return keymap[1][i];	//second row pressed
			case 0xFD:	return keymap[2][i];	//third row pressed
			case 0xFE:	return keymap[3][i];	//fourth row pressed
			default:	break;					//no pressed button found on this column
		}
	}
	return 16;	//no button pressed => blank 7 segment
}

void printf( char *data){
	while(*data)
	{
		while(BusyUSART());
		putcUSART(*data);
		*data++;
	}
}

void main(void)
{
	init();					//initialization

	while(1){
		
		input = scan();		//scan the keypad for pressed buttons

		RestartI2C();		//restart I2C communication
			IdleI2C();		//wait until I2C is ready
		WriteI2C( 0x42 );	//send address to the device (0100 (hardcoded in chip) + 001 (address pins) + 0 (write mode))
			IdleI2C();		//wait until I2C is ready
		WriteI2C( ~sevSeg[input] );	//send input to the 7 segment display (the input is first converted to its 7 segment code
									//and then inverted because the 7 segment is common anode)
			IdleI2C();		//wait until I2C is ready

		if(input != 16 && input != previnput){	//print the input once when a button is pressed
			itoa(input,str+7);	//convert the input to a string and store after "Input: "
			strcat(str,end);	//append newline
			printf(str);		//write string to UART
			while(BusyUSART());	//wait until I2C is ready
		}
		previnput = input;	//update the previous input to the current input
	}
}

